﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Windows.Forms;

namespace ScriptCaching
{
    public partial class Form1 : Form
    {
        private string root_;
        private ScriptManager script_manager_;
        private Dictionary<string, TreeNode> tree_nodes_ = new Dictionary<string, TreeNode>();

        public Form1()
        {
            InitializeComponent();
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            // locate the scripts
            root_ = typeof(Form1).Assembly.Location;
            while (!Directory.Exists(root_ + @"\Nodes"))
                root_ = Path.GetDirectoryName(root_);

            root_ += @"\Nodes";

            // initialize the script manager
            script_manager_ = new ScriptManager(basicIdeCtl1, root_);

            // script manager event for updating list view as script status changes
            script_manager_.ScriptLoadedStatusChanged += (script_name, script_loaded_status) =>
            {
                TreeNode tree_nodex = null;
                if (tree_nodes_.TryGetValue(script_name, out tree_nodex))
                    tree_nodex.Text = tree_nodex.Text.Substring(0, tree_nodex.Text.Length - 1) + (script_loaded_status ? "+" : " ");
            };

            // initialize the tree view and script nodes
            var tree_node = new TreeNode(@"\");
            tree_node.Tag = "<none>";
            treeView1.Nodes.Add(tree_node);
            InitializeTree(new DirectoryInfo(root_), tree_node);
            treeView1.ExpandAll();
            treeView1.SelectedNode = tree_node;
        }

        private void Form1_FormClosed(object sender, FormClosedEventArgs e)
        {
            // free script manager resources
            script_manager_.Dispose();
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            // prevent close if the script manager can't shut down
            e.Cancel = !script_manager_.Shutdown();
        }

        private void basicIdeCtl1_EnterDesignMode(object sender, WinWrap.Basic.Classic.DesignModeEventArgs e)
        {
            // disable the run script button
            btnRunScripts.Enabled = false;
            checkBox1.Enabled = true;
        }

        private void basicIdeCtl1_LeaveDesignMode(object sender, WinWrap.Basic.Classic.DesignModeEventArgs e)
        {
            // enable the run script button
            btnRunScripts.Enabled = true;
            checkBox1.Enabled = false;
        }

        private void basicIdeCtl1_OverrideMenuCommand(object sender, WinWrap.Basic.Classic.OverrideMenuCommandEventArgs e)
        {
            switch (e.CmdId)
            {
                case WinWrap.Basic.CommandConstants.FileClose:
                case WinWrap.Basic.CommandConstants.FileOpen:
                case WinWrap.Basic.CommandConstants.FileSaveAs:
                case WinWrap.Basic.CommandConstants.NewModulePopup:
                    // these commands are not needed in this sample application
                    e.Cancel = true;
                    break;
            }
        }

        private void treeView1_AfterSelect(object sender, TreeViewEventArgs e)
        {
            var script_name = (string)e.Node.Tag;
            if (script_name != null)
            {
                // view code into the IDE
                basicIdeCtl1.FileName = script_name;
            }
            else
            {
                // no code to view
                basicIdeCtl1.FileName = "<none>";
            }
        }

        private void btnRunScripts_Click(object sender, EventArgs e)
        {
            // clear the immediate window
            using (var q = basicIdeCtl1.Query("Set IDE Immediate", "")) { }
            // run all the script in the sub tree
            var script_names = ScriptNames((TreeNode)treeView1.SelectedNode);
            var results = script_manager_.EvaluateScripts(script_names, "<null>");
            // show the results in the immediate window
            using (var q = basicIdeCtl1.Query("Set IDE Immediate", string.Join("\r\n", results))) { }
        }

        private void checkBox1_Click(object sender, EventArgs e)
        {
            script_manager_.Fast = checkBox1.Checked;
        }

        private void InitializeTree(DirectoryInfo parent_di, TreeNode parent_tree_node)
        {
            // recursively initialize the tree from the parent folder and sub-folders
            foreach (var fi in parent_di.GetFiles())
            {
                var child_tree_node = parent_tree_node.Nodes.Add(fi.Name + " "); // not loaded
                string script_name = fi.FullName.Substring(root_.Length);
                child_tree_node.Tag = script_name; // not loaded
                tree_nodes_.Add(script_name, child_tree_node);
            }

            foreach (var child_di in parent_di.GetDirectories())
            {
                var child_tree_node = parent_tree_node.Nodes.Add(child_di.Name + @"\");
                child_tree_node.Tag = "<none>";
                InitializeTree(child_di, child_tree_node);
            }
        }

        private List<string> ScriptNames(TreeNode tree_node)
        {
            // return the list of script nodes at and under this tree node
            var script_names = new List<string>();
            BuildScriptNodeList(tree_node, script_names);
            return script_names;
        }

        private void BuildScriptNodeList(TreeNode tree_node, List<string> script_names)
        {
            // recursive for one tree node
            var script_name = (string)tree_node.Tag;
            if (script_name == "<none>")
                BuildChildrenScriptNodeList(tree_node, script_names);
            else
                script_names.Add(script_name);
        }

        private void BuildChildrenScriptNodeList(TreeNode parent_tree_node, List<string> script_names)
        {
            // recursive for all child tree nodes
            foreach (TreeNode child_tree_node in parent_tree_node.Nodes)
                BuildScriptNodeList(child_tree_node, script_names);
        }
    }
}
