﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace ScriptCaching
{
    // used by WinWrap Basic to read/write script to the virtual file system
    // in this case the real file system
    internal class ScriptManager : WinWrap.Basic.IVirtualFileSystem, IDisposable
    {
        // notify subscribers of script loaded status change
        public event Action<string, bool> ScriptLoadedStatusChanged;

        // this basic object will load and execute the scripts
        private WinWrap.Basic.BasicNoUIObj basicNoUIObj_;
        // root directory for all scripts
        private string root_;
        // simple extension for evaluating scripts within scripts
        private CExtensions extensions_;

        // max number of scripts loaded (cached)
        private const int max_loaded_scripts = 3; // arbitrarily low

        // this is the dictionary of all the scripts
        private List<ScriptNode> loaded_script_nodes_ = new List<ScriptNode>();
        ScriptNode active_script_node_;

        // use the Fast interface instead of the Handler interface to evaluate scripts
        public bool Fast { get; set; }

        public ScriptManager(WinWrap.Basic.BasicIdeCtl basicIdeCtl, string root)
        {
            root_ = root;
            // create the language extensions
            extensions_ = new CExtensions(this);

            // create the NoUI script engine
            basicNoUIObj_ = new WinWrap.Basic.BasicNoUIObj();
            // set the secret (not needed when running under a Development Certificate)
            basicNoUIObj_.Secret = new Guid("00000000-0000-0000-0000-000000000000");
            // initialize
            basicNoUIObj_.Initialize();
            // use design mode
            basicNoUIObj_.DesignModeVisible = true;

            // synchronize the IDE with the NoUI 
            basicNoUIObj_.Synchronizing += (sender, e) => basicIdeCtl.Synchronize(e.Param, e.Id);
            // synchronize the NoUI with the IDE
            basicIdeCtl.Synchronizing += (sender, e) => basicNoUIObj_.Synchronize(e.Param, e.Id);
            // allow the NoUI to be synchronized for editing
            basicNoUIObj_.SynchronizedEdit = true;
            // cause the IDE to be synchronized for editing
            basicIdeCtl.SynchronizedEdit = true;

            // implement DoEvents
            basicNoUIObj_.DoEvents += (sender, e) => WinWrap.Basic.Util.DoEvents();

            // unload all previously loaded scripts
            basicNoUIObj_.EnterDesignMode += (sender, e) =>
            {
                if (active_script_node_ != null)
                    e.Cancel = true; // executing in a script, can't unload
                else
                {
                    // unload every loaded script
                    foreach (var script_node in loaded_script_nodes_)
                    {
                        script_node.Unload();
                        ScriptLoadedStatusChanged?.Invoke(script_node.ScriptName, false);
                    }

                    loaded_script_nodes_.Clear();

                    // collect all Script object
                    GC.Collect();
                    GC.WaitForPendingFinalizers();
                }
            };

            // use WWB-COM language with CreateHandler (needed for a .NET application running under a Development Certificate)
            basicNoUIObj_.AddExtension("$Feature WWB-COM True", null);

            // .NET 5 does not create a type library or type info for COM interop
            // use type library created by .NET 4
            basicNoUIObj_.AddExtension(@"{}<APPDIR>\ScriptCaching-Net5.tlb", null);

            // add the language extensions, specify type because .NET 5 does not implement IDispatch::GetTypeInfo
            basicNoUIObj_.AddExtension("(Extensions)", extensions_);

            // WinWrap Basic reads/writes scripts using this class
            basicNoUIObj_.VirtualFileSystem = this;
        }

        public object EvaluateScript(ScriptNode script_node, object default_value = null)
        {
            // script is active
            ScriptNode last_active_script_node = active_script_node_;
            active_script_node_ = script_node;
            try
            {
                // a script run-time error will throw an exception
                return active_script_node_.Evaluate(default_value);
            }
            finally
            {
                // script is no longer active
                active_script_node_ = last_active_script_node;
            }
        }

        public List<object> EvaluateScripts(List<string> script_names, object default_value = null)
        {
            var results = new List<object>();
            try
            {
                // evaluate each script in order
                foreach (var script_name in script_names)
                {
                    // load the script, a script syntax error throws an exception
                    var script_node = LoadScript(script_name);
                    // evaluate the script, a script run-time error throws an exception
                    results.Add(EvaluateScript(script_node, default_value));
                }
            }
            catch (WinWrap.Basic.TerminatedException)
            {
                // special exception for user termination of a script
                results.Add("<execution terminated>");
            }
            catch (Exception ex)
            {
                // report the syntax or run-time error to the remote
                basicNoUIObj_.ReportError(ex);
                // add the exception to the results
                results.Add($"<exception: {ex.Message.Replace("\r\n", " -- ")}>");
            }

            // always returns result, never throws an exception
            return results;
        }

        // return true if no scripts are running
        public bool Shutdown() => basicNoUIObj_.Shutdown() == 0;

        public ScriptNode LoadScript(string script_name)
        {
            if (!script_name.StartsWith(@"\"))
            {
                // relative script path
                if (active_script_node_ != null)
                    script_name = Combine(active_script_node_.ScriptName, script_name);
                else
                    script_name = @"\" + script_name;
            }

            // is the script already loaded
            ScriptNode script_node = loaded_script_nodes_.Find(script_nodex => script_nodex.ScriptName == script_name);
            if (script_node == null)
                // not loaded
                script_node = new ScriptNode(script_name);

            // evaluate a script node
            if (!script_node.IsLoaded)
            {
                // load the script, a syntax error will throw an exception
                script_node.Load(basicNoUIObj_, Fast);
                // add to the front of the loaded list
                loaded_script_nodes_.Insert(0, script_node);
                // update tree view with new status
                ScriptLoadedStatusChanged?.Invoke(script_name, true);

                while (loaded_script_nodes_.Count > max_loaded_scripts)
                {
                    // remove the last (oldest loaded script node)
                    var oldest_script_node = loaded_script_nodes_[loaded_script_nodes_.Count-1];
                    if (oldest_script_node.IsActive)
                        break;

                    loaded_script_nodes_.RemoveAt(loaded_script_nodes_.Count - 1);
                    oldest_script_node.Unload();
                    // update tree view with new status
                    ScriptLoadedStatusChanged?.Invoke(oldest_script_node.ScriptName, false);
                }
            }
            else if (script_node != loaded_script_nodes_[0])
            {
                // move to the front of the loaded list
                loaded_script_nodes_.Remove(script_node);
                loaded_script_nodes_.Insert(0, script_node);
            }

            return script_node;
        }

        private string Combine(string baseFileName, string name)
        {
            if (name == "<none>")
                return name;

            // determine the root directory
            string root = string.IsNullOrEmpty(baseFileName) ? root_ : Path.GetDirectoryName(root_ + baseFileName);

            // script names that start with \ are absolute paths
            if (name.StartsWith(@"\"))
                name = name.Substring(1);

            // create a path
            name = Path.Combine(root, name);
            // use a full path
            name = Path.GetFullPath(name);
            // insure that the name is still within the one of the root_ sub directories
            if (!name.StartsWith(root_ + @"\"))
                return "<none>";

            // this is the script name
            return name.Substring(root_.Length);
        }

        #region IVirtualFileSystem
        // http://www.winwrap.com/web2/solution/bp-VirtualFileSystem
        string WinWrap.Basic.IVirtualFileSystem.Combine(string baseFileName, string name) => Combine(baseFileName, name);
        void WinWrap.Basic.IVirtualFileSystem.Delete(string scriptPath) => throw new NotImplementedException();
        bool WinWrap.Basic.IVirtualFileSystem.Exists(string scriptPath) => File.Exists(root_ + scriptPath);
        string WinWrap.Basic.IVirtualFileSystem.GetCaption(string scriptPath) => scriptPath != "<none>" ? Path.GetFileNameWithoutExtension(scriptPath) : "<none>";
        DateTime WinWrap.Basic.IVirtualFileSystem.GetTimeStamp(string scriptPath) => throw new NotImplementedException();
        string WinWrap.Basic.IVirtualFileSystem.Read(string scriptPath) => scriptPath != "<none>" ? File.ReadAllText(root_ + scriptPath) : "'#ReadOnly";
        void WinWrap.Basic.IVirtualFileSystem.Write(string scriptPath, string text) => File.WriteAllText(root_ + scriptPath, text, Encoding.UTF8);
        #endregion

        public void Dispose() => basicNoUIObj_.Dispose();
    }
}
